// Floppy Controller DP8473
// Start 16.4.2023

module DP8473 (BCLK, CERES_ON, WSEL, RSEL, ADDR, DIN, SSW, SATA_Q, DP8473_Q, INTERRUPT, STATUS, CER_SATA);

	input			BCLK;
	input			CERES_ON;
	input			WSEL;
	input			RSEL;
	input	 [2:0]	ADDR;
	input	 [7:0]	DIN;
	input	 [3:0]	SSW;
	input	[64:0]	SATA_Q;
	
	output	reg  [7:0]	DP8473_Q;
	output	reg		INTERRUPT;
	output	reg  [7:0]	STATUS;
	output	[37:0]	CER_SATA;

	reg				read;
	reg				reset;	// active high
	reg		 [1:0]	state;
	reg				old_com;
	reg		 [7:0]	command;
	reg		 [3:0]	para_w;
	reg		 [2:0]	para_r;
	reg		 [8:0]	lenge;
	reg		 [7:0]	param_0,param_1,param_2,param_3,param_4,param_5;
	reg				etr,ips;	// etr = extended track - gebraucht ?
	reg		 [7:0]	response;
	reg		 [6:0]	track;
	reg		 [8:0]	s_state;
	reg		[11:0]	anz;
	reg		[31:0]	data_reg;
	reg		 [2:0]	banz;
	reg		 [6:0]	fanz;
	reg				error;
	reg		 [1:0]	floppy;
	reg		 [3:0]	tasten;
	reg		[35:0]	cer_data;
	reg				wr_led,do_flush,flush,fl_ende,fl_wr,fl_kom;	// alles fuer FLUSH
	reg		 [3:0]	f_state;
	
	wire			write;
	wire			ini_phase,com_phase,exe_phase,res_phase;
	wire			load;
	wire			eow,eor;
	wire			wr_param;
	wire			done;
	wire			req_master;
	wire			go_read,go_sata,format;
	wire			get_data,put_data;
	wire			full_tr;
	wire	[10:0]	lba;	// beschraenkt auf 720 KB Disks
	wire	 [4:0]	sec;	// kann 18 Sektoren fuer 1,44 MB Disks
	wire			anz_done;
	wire			byte_done;
	wire			fl_run;
	
	assign write = WSEL & (ADDR == 3'd5);
	always @(posedge BCLK) read <= RSEL & (ADDR == 3'd5);	// 1 Takt spaeter - nur 1 Takt breit !

	always @(posedge BCLK) if (WSEL && (ADDR == 3'd2)) reset <= ~DIN[2];	// Drive Control Register, nur Bit 2 fuer Reset, Rest Motor Control

	always @(posedge BCLK)
		if (reset) state <= 2'd0;
		  else
			casex ({write,eow,done,eor,state})
			  6'b0x_x_x_00 : state <= 2'd0;
			  6'b1x_x_x_00 : state <= 2'd1;
			  6'bx0_x_x_01 : state <= 2'd1;
			  6'bx1_0_x_01 : state <= 2'd2;
			  6'bx1_1_0_01 : state <= 2'd3;
			  6'bx1_1_1_01 : state <= 2'd0;
			  6'bxx_0_x_10 : state <= 2'd2;
			  6'bxx_1_x_10 : state <= 2'd3;
			  6'bxx_x_0_11 : state <= 2'd3;
			  6'bxx_x_1_11 : state <= 2'd0;
			  default	   : state <= 2'd0;
			endcase
	
	assign ini_phase = (state == 2'd0);
	assign com_phase = (state == 2'd1);
	assign exe_phase = (state == 2'd2);
	assign res_phase = (state == 2'd3);

	always @(posedge BCLK) old_com <= com_phase;
	assign load = ~com_phase & old_com;
	
	always @(posedge BCLK) if (write && ini_phase) command <= DIN;
	
	always @(posedge BCLK)
		if (write && ini_phase) para_w <= 4'd0;
		  else
			if (write && com_phase) para_w <= para_w + 4'd1;
	
	always @(posedge BCLK)
		if (write && ini_phase) para_r <= 3'd0;
		  else
			if (read && res_phase) para_r <= para_r + 3'd1;

	always @(command or etr)
		casex (command)
		  8'bxxx0_0110 : lenge =  9'h187;	// Read Data
		  8'b0x00_1010 : lenge =  9'h01F;	// Read ID
		  8'b0x00_1101 : lenge =  9'h157;	// Format
		  8'bxxx0_1100 : lenge =  9'h08F;	// Read Deleted Data
		  8'b0xx0_0010 : lenge =  9'h187;	// Read a Track
		  8'bxx00_0101 : lenge =  9'h187;	// Write Data
		  8'bxx00_1001 : lenge =  9'h08F;	// Write Deleted Data
		  8'bxxx1_0001 : lenge =  9'h08F;	// Scan Equal
		  8'bxxx1_1001 : lenge =  9'h08F;	// Scan Low or Equal
		  8'bxxx1_1101 : lenge =  9'h08F;	// Scan High or Equal
		  8'b0000_1111 : lenge = {1'b0,(etr ? 4'd3 : 4'd2),4'd0};	// Seek
		  8'b0000_0111 : lenge =  9'h010;	// Recalibrate
		  8'b0000_1000 : lenge = {5'd0,(etr ? 4'd3 : 4'd2)};		// Sense Interrupt
		  8'b0000_0100 : lenge =  9'h011;	// Sense Drive Status
		  8'b0000_0011 : lenge =  9'h020;	// Specify
		  8'b0000_0001 : lenge =  9'h040;	// Mode
		  8'b0x10_0001 : lenge =  9'h029;	// Set Track - Spezial Kommando, wird sicher nicht gebraucht...
		  default	   : lenge =  9'd008;	// Invalid Command : Bit 3 = 1
		endcase
		
	assign eow = (lenge[7:4] == para_w);
	assign eor = (lenge[2:0] == para_r);
			
	assign wr_param = write & com_phase;
																				// fur READ,READ A TRACK,WRITE & FORMAT
	always @(posedge BCLK) if (wr_param && (para_w == 4'd0)) param_0 <= DIN;	// IPS, Head & Drive
	always @(posedge BCLK) if (wr_param && (para_w == 4'd1)) param_1 <= DIN;	// Track
	always @(posedge BCLK) if (wr_param && (para_w == 4'd2)) param_2 <= DIN;	// Head
	always @(posedge BCLK) if (wr_param && (para_w == 4'd3)) param_3 <= DIN;	// Sector
	always @(posedge BCLK) if (wr_param && (para_w == 4'd4)) param_4 <= DIN;	// Bytes in Sector
	always @(posedge BCLK) if (wr_param && (para_w == 4'd5)) param_5 <= DIN;	// End of Track Sector Number

	always @(posedge BCLK)
		if (reset) etr <= 1'b0;	// Extended Track Range - wird nicht gesetzt in Ceres-3
		  else
			if (load && (command == 8'd1)) etr <= param_0[0];	// Mode
			
	always @(posedge BCLK)
		if (reset) ips <= 1'b0;	// Implied Seek
		  else
			if (load && (command == 8'd1)) ips <= param_0[5];	// Mode
			
	always @(posedge BCLK)
		if (load && (command == 8'd7)) track <= 7'd0;	// Recalibrate
		  else
			if (load && ((command == 8'h0F) || (param_0[7] && ips && lenge[8]))) track <= param_1[6:0];	// Seek oder alle anderen Kommandos
			
	assign done = (s_state[3:0] == 4'hF) | ~lenge[8];
	
	// wann ist "Request for Master" ?
	assign req_master = ini_phase | com_phase | res_phase | s_state[8];
	
	always @(*)
		casex ({exe_phase,para_r})
		  4'd0    : response = {lenge[3],error,6'h20};	// Status 0 : Seek End
		  4'd1    : response = (command == 8'd8) ? {1'b0,track} : 8'd0;	// Status 1 oder Tracknummer fuer Sense Interrupt
		  4'd2	  : response = 8'd0;	// Status 2
		  4'd3	  : response = track;	// Track, kein Update wie im Datenblatt beschrieben !
		  4'd4	  : response = {7'd0,param_2[0]};	// Head
		  4'd5	  : response = {3'd0,param_3[4:0]};	// Sector
		  4'd6	  : response = {5'd0,param_4[2:0]};	// Bytes in Sector
		  4'd7    : response = 8'hxx;	// egal
		  default : response = data_reg[7:0];	// Daten
		endcase
		
	always @(posedge BCLK) DP8473_Q <= ADDR[0] ? response							// 4 LSB : kein Seeking of Drive  --v
											   : {req_master,(res_phase | (exe_phase & go_read)),exe_phase,~ini_phase,4'd0};
	
	always @(posedge BCLK) INTERRUPT <= 1'b1;
	
	// +++++++++++++++ Kommandos verarbeiten +++++++++++++

	assign go_read = (command[4:0] == 5'd6) | (command[4:0] == 5'd2);	// READ oder READ A TRACK
	assign go_sata = load & lenge[8] & (go_read | (command[4:0] == 5'd5) | (command[4:0] == 5'd13));	// WRITE oder FORMAT
	assign format  = command[3];
	
	always @(posedge BCLK)
		if (reset) s_state <= 9'd0;
		  else
			casex ({go_sata,flush,go_read,format,SATA_Q[64],SATA_Q[31],byte_done,anz_done,s_state[4:0]})
			  13'b0xxx_xxxx_00000 : s_state <= 9'b0000_00000;	// nix zu tun
			  13'b1xxx_xxxx_00000 : s_state <= 9'b0000_00001;	// jetzt geht es los
			  13'bx1xx_xxxx_00001 : s_state <= 9'b0000_00001;	// FLUSH laeuft ! Warten...
			  13'bx0xx_xxxx_00001 : s_state <= 9'b0001_00010;	// Fertig. Weiter.
			  13'bxxxx_xxxx_00010 : s_state <= 9'b0001_00011;	// LBA schreiben
			  13'bxxxx_xxxx_00011 : s_state <= 9'b0001_00100;	// Sektor Anzahl
			  13'bxxxx_xxxx_00100 : s_state <= 9'b0000_00101;	// Kommando raus
			  13'bxxxx_xxxx_00101 : s_state <= 9'b0000_00110;	// jetzt ein bisschen warten
  			  13'bxxxx_xxxx_00110 : s_state <= 9'b0000_00111;
  			  13'bxx1x_xxxx_00111 : s_state <= 9'b0010_01000;	// zum Lesen
			  13'bxx00_xxxx_00111 : s_state <= 9'b1000_01010;	// zum Schreiben
			  13'bxx01_xxxx_00111 : s_state <= 9'b1100_11000;	// zum Format : Master Request und SATA Write aktiv
			  // *** Daten READ
			  13'bxxxx_0xxx_01000 : s_state <= 9'b0010_01000;	// warten auf READY
			  13'bxxxx_1xxx_01000 : s_state <= 9'b1000_01001;
			  13'bxxxx_xx0x_01001 : s_state <= 9'b1000_01001;
			  13'bxxxx_xx10_01001 : s_state <= 9'b0010_01000;	// weiter
			  13'bxxxx_xx11_01001 : s_state <= 9'b0000_01100;	// warten auf INT
			  // *** Daten WRITE
			  13'bxxxx_xx0x_01010 : s_state <= 9'b1000_01010;	// warten auf 4 Bytes geschrieben
			  13'bxxxx_xx1x_01010 : s_state <= 9'b0100_01011;
			  13'bxxxx_0xxx_01011 : s_state <= 9'b0100_01011;	// warten auf READY
			  13'bxxxx_1xxx_01011 : s_state <= 9'b0000_01110;	// zum Extratakt fuer neuen anz Wert
			  13'bxxxx_xxx0_01110 : s_state <= 9'b1000_01010;
			  13'bxxxx_xxx1_01110 : s_state <= 9'b0000_01100;	// warten auf INT
			  // *** Format - beide Operationen parallel !
			  13'bxxxx_0x0x_11000 : s_state <= 9'b1100_11000;	// alle 4 Moeglichkeiten
			  13'bxxxx_1x0x_11000 : s_state <= 9'b1000_11001;
			  13'bxxxx_1x1x_11000 : s_state <= 9'b0000_11100;
			  13'bxxxx_0x1x_11000 : s_state <= 9'b0100_11011;
			  //
			  13'bxxxx_xx00_11001 : s_state <= 9'b1100_11000;	// alle 4 Moeglichkeiten
			  13'bxxxx_xx01_11001 : s_state <= 9'b1000_11010;
			  13'bxxxx_xx10_11001 : s_state <= 9'b0100_11011;
			  13'bxxxx_xx11_11001 : s_state <= 9'b0000_01100;	// warten auf INT
			  // nur CPU schreibt
			  13'bxxxx_xx0x_11010 : s_state <= 9'b1000_11010;
			  13'bxxxx_xx1x_11010 : s_state <= 9'b0000_01100;	// warten auf INT
			  // nur SATA Schreiben :
			  13'bxxxx_0xxx_11011 : s_state <= 9'b0100_11011;
			  13'bxxxx_1xxx_11011 : s_state <= 9'b0000_11100;
			  13'bxxxx_xxx0_11100 : s_state <= 9'b0100_11011;
			  13'bxxxx_xxx1_11100 : s_state <= 9'b0000_01100;	// warten auf INT
			  // *** Warten auf Ende: INT=1, state[1:0] == 00 !
			  13'bxxxx_x0xx_01100 : s_state <= 9'b0000_01100;
			  13'bxxxx_x1xx_01100 : s_state <= 9'b0000_01111;	// Ende mit s_state[3:0] == 4'hF !
			  13'bxxxx_xxxx_01111 : s_state <= 9'b0000_00000;
			  default			  : s_state <= 9'd0;
			endcase
			
	assign get_data = s_state[6] & SATA_Q[64];
	assign put_data = s_state[7] & SATA_Q[64];
	
	assign full_tr = (command[4:0] == 5'd2) | (command[4:0] == 5'd13);	// READ A TRACK oder FORMAT
	//			 Track * 18							+ Head							+ Sektor		, zur Zeit 11 Bit fuer 720 KB
	assign lba = {track,4'd0} + {3'd0,track,1'b0} + (param_0[2] ? 11'd9 : 11'd0) + (full_tr ? 11'd0 : {6'd0,(param_3[4:0] - 5'd1)});
	assign sec = format ? param_2[4:0] : (param_5[4:0] - ((command[4:0] == 5'd2) ? 5'd0 : (param_3[4:0] - 5'd1)));
			
	always @(posedge BCLK)
		if (s_state[5]) anz <= {sec,7'd0};	// Double Worte 
		  else
			if (get_data || put_data) anz <= anz - 12'd1;	// fuer Lesen und Schreiben

	assign anz_done = (anz == 12'd0);
	
	always @(posedge BCLK)
		if (get_data) data_reg <= SATA_Q[63:32];	// Lesedaten...
		  else
			if (read || write) data_reg <= {DIN,data_reg[31:8]};	// ...und Schreibdaten
			
	always @(posedge BCLK)
		if (get_data || put_data || s_state[5]) banz <= 3'd4;	// Byte Anzahl initialisieren
		  else
			if (read || write) banz <= banz - 3'd1;
			
	always @(posedge BCLK)
		if (s_state[5]) fanz <= {sec,2'd0};	// Byte Anzahl initialisieren fuer Format
		  else
			if (write) fanz <= fanz - 7'd1;
			
	assign byte_done = format ? (fanz == 7'd0) : (banz == 3'd0);
	
	always @(posedge BCLK) error = (SATA_Q[15:0] != 16'h0050) & lenge[8];
	
	always @(posedge BCLK)
		if (flush) cer_data[31:0] <= {24'd0,(fl_kom ? 8'hE7 : 8'd0)};
		  else
			casex ({s_state[5],s_state[1:0]})
			  3'b1_10 : cer_data[31:0] <= {16'h0200,2'd0,floppy,1'b0,lba};	// je 2 MByte
			  3'b1_11 : cer_data[31:0] <= {27'd0,sec};
			  3'b1_0x : cer_data[31:0] <= {24'd0,(go_read ? 8'hC8 : 8'hCA)};
			  default : cer_data[31:0] <= format ? {4{param_4}} : data_reg;
			endcase
		
	always @(posedge BCLK) cer_data[35:34] <= {CERES_ON,(s_state[5] | fl_wr)};	// WRITE in SATA;
	always @(posedge BCLK) cer_data[33:32] <= flush ? f_state[1:0] : (s_state[3] ? 2'd0 : s_state[1:0]);	// Adresse nur 2 Bits
	
	assign CER_SATA = {cer_data[35],s_state[7:6],cer_data[34:0]};	// s_state fuer schnelles RDY Response
	
	// +++++++++++++++ Floppy Auswahl und Status +++++++++++++
			
	always @(posedge BCLK) tasten <= ~SSW;	// ohne Register macht Bloedsinn bei Taste 1 ...
	
	always @(posedge BCLK)
		if (CERES_ON && !cer_data[35]) floppy <= 2'd0;
		  else
			if (tasten != 4'd0) floppy <= {(tasten[3] | tasten[2]),(tasten[3] | tasten[1])};
			
	always @(posedge BCLK) wr_led <= CERES_ON & (s_state[7] | (wr_led & ~flush));
			
	always @(posedge BCLK) STATUS[7] <= (floppy == 2'd3);
	always @(posedge BCLK) STATUS[6] <= STATUS[7] & wr_led;
	always @(posedge BCLK) STATUS[5] <= (floppy == 2'd2);
	always @(posedge BCLK) STATUS[4] <= STATUS[5] & wr_led;
	always @(posedge BCLK) STATUS[3] <= (floppy == 2'd1);
	always @(posedge BCLK) STATUS[2] <= STATUS[3] & wr_led;
	always @(posedge BCLK) STATUS[1] <= (floppy == 2'd0);
	always @(posedge BCLK) STATUS[0] <= STATUS[1] & wr_led;
	
	assign fl_run = CERES_ON & ~reset;
	
	always @(posedge BCLK) do_flush <= fl_run & (((tasten != 4'd0) & wr_led) | (do_flush & ~flush));

	always @(posedge BCLK) flush <= fl_run & ((do_flush &	ini_phase) | (flush & ~fl_ende));
	
	always @(posedge BCLK) f_state <= fl_run ? f_state + {3'd0,(flush & ~f_state[3])} : 4'd0;
	
	always @(posedge BCLK) fl_wr  = (f_state[3:2] == 2'd0) & (f_state[1:0] != 2'd0);
	always @(posedge BCLK) fl_kom = (f_state[3:0] == 4'd3);
	
	always @(posedge BCLK) fl_ende <= f_state[3] & SATA_Q[31] & (SATA_Q[15:0] == 16'h0050);
	
//	always @(posedge BCLK) STATUS <= {error,anz_done,state,s_state[3:0]};
	
endmodule
