// 15.10.2023 for SoC-DE1

module MAINDEC
	(BCLK, RST_N, SSW, WRITE, READ, STATUS, ADDR, BE, DIN, BRESET, ENDRAM, DRAMSZ, WRSEL, RDSEL, CWSEL, CRSEL,
	 BOOT_Q, CFG_Q, CERES_Q, PS2_Q, OBERON_Q, OROM_Q, OROM_RDY, OROM,
	 NMI_N, READY, DRD, LEDR, SSEG, CONFIG, HPS_IN, HPS_Q, HPS_OUT );
	
	input 			BCLK;
	input			RST_N;
	input	[12:0]	SSW;	// 10..12 sind die Keys
	input 			WRITE;
	input 			READ;
	input	 [3:0]	STATUS;
	input	[31:2]	ADDR;
	input	 [3:0]	BE;
	input	[31:0]	DIN;
	
	output	reg		BRESET;
	output	reg		ENDRAM;		// 21.11.2014
	output	 [2:0]	DRAMSZ;

	output	[15:0]	WRSEL,RDSEL;
	output	[15:8]	CWSEL,CRSEL;
	
	input   [31:0]	BOOT_Q;		// Datenbusse
	input    [8:0]	CFG_Q;	
	input	[31:0]	CERES_Q;
	input	[15:0]	PS2_Q;
	input	[31:0]	OBERON_Q;
	
	input	[31:0]	OROM_Q;
	input			OROM_RDY;
	output			OROM;
	
	output	reg		NMI_N;

	output			READY;
	output  reg	[31:0]	DRD;

	output	reg	 [41:0]	SSEG;
	output	reg	 [9:0]	LEDR;
	
	output	 [4:0]	CONFIG;
	
	input	[31:0]	HPS_IN,HPS_Q;
	output	[31:0]	HPS_OUT;
	
	reg		[15:0]	select;
	reg		[31:0]	rdmux;
	reg				rd_rdy;
	reg		 [3:0]	init_cou;
	reg		[31:0]	counter;
	reg				nmie;
	reg		[41:0]	ss_reg;
	reg		[19:0]	halt_cnt;
	reg				inc_halt_cnt;
	reg				ceres,obe_ver,dual_mode,hex_out;	// = CONFIG
	reg		 [9:0]	tick_c;
	reg				nmi_puls;
	reg				old_c15;
	reg		[31:0]	hps_reg;

	wire			wr_coun;
	wire			access;
	wire			second,tick,shift,sec_puls;	// Sekundentimer
	
	// +++++++++++++++ Debug +++++++++++++
	
	assign access  = WRITE | READ;
	
	always @(ADDR)
		case (ADDR[15:12])
			4'h0 : select = 16'h0001;	// Boot ROM
			4'h1 : select = 16'h0002;	// Boot ROM
			4'h2 : select = 16'h0004;	// UART
			4'h3 : select = 16'h0008;	// Configurationsspeicher
			4'h4 : select = 16'h0010;	// I2C
			4'h5 : select = 16'h0020;	// NMI Counter, LEDs, lesen des SSW und der Statistik
			4'h6 : select = 16'h0040;	// Grafik + DMA
			4'h7 : select = 16'h0080;	// Terminal
			4'h8 : select = 16'h0100;	// SDC PC532 : Control
			4'h9 : select = 16'h0200;	// SDC : PERI
			4'hA : select = 16'h0400;	// Ethernet : PERI
			4'hB : select = 16'h0800;	// PS2 : PERI
			4'hC : select = 16'h1000;	// SATA : PERI
			4'hD : select = 16'h2000;	// Audio
			4'hE : select = 16'h4000;	// CAM
			4'hF : select = 16'h8000;	// Modulauswahl : PC532, etc.
		endcase

	// Lesen aus Block-RAM erfordert einen Warte-Takt, daher alle Lese-Zugriffe 2 Takte
	always @(posedge BCLK) rd_rdy <= READ & (OROM ? OROM_RDY : ~rd_rdy) ;	
	
	// READ Puls als 1. Lesestufe in den Peripherie Bloecken
	assign RDSEL = READ  & ~rd_rdy & (ADDR[31:20] == 12'hE00) & ~ceres ? select : 16'd0;
	assign WRSEL = WRITE & 			 (ADDR[31:20] == 12'hE00) & ~ceres ? select : 16'd0;

	assign CRSEL = READ  & ~rd_rdy & (ADDR[31:20] == 12'hFFF) & BE[0]  ? select[15:8] : 8'd0;
	assign CWSEL = WRITE &           (ADDR[31:20] == 12'hFFF) & BE[0]  ? select[15:8] : 8'd0;
	
	assign OROM  = READ  & (|SSW[2:1]) & (ADDR[31:27] == 5'b1111_0);
	
	assign READY =  rd_rdy | WRITE ;
	
	// 2. Level READ Daten Multiplexer
	always @(*)
	  casex({ADDR[15:12]})
	    4'b000x : rdmux = BOOT_Q;	// Boot-ROM
		4'h3 	: rdmux = {23'd0,CFG_Q};
		4'h5 	: rdmux = (ADDR[11:9] == 3'd0) ? ((ADDR[3:2] == 2'd3) ? {nmie,1'b0,hex_out,29'd0} : counter)
											   : {19'd0,SSW};
		4'hC    : rdmux = HPS_IN;
		4'hD    : rdmux = HPS_Q;
		4'hB	: rdmux = {16'd0,PS2_Q};	// Port fuer Debug wenn noetig
		default : rdmux = 32'd0;
	  endcase
	  
	// 1. Level READ Daten Multiplxer
	always @(*)
	  casex({ADDR[31:24]})
	    8'h0x	: DRD = BOOT_Q;		// zum Starten !
	    8'hE0   : DRD = rdmux;
		8'b1111_0xxx : DRD = |SSW[2:1] ? OROM_Q : OBERON_Q;
		8'b1111_1xxx : DRD = CERES_Q;
		default : DRD = 32'd0;
	  endcase
	
	// Programmzugriff im oberen ROM Bereich schaltet DRAM ein
	always @(posedge BCLK) ENDRAM <= (((ADDR[31:28] == 4'hE) & (STATUS == 4'h8) & READ) | ENDRAM) & BRESET;

	// ++++++++++++++++++++++++++ HPS Register & Mux ++++++++++++++++++++++++++
	
	always @(posedge BCLK) if (WRSEL[12]) hps_reg <= DIN;
	
	assign HPS_OUT = ceres ? HPS_Q : hps_reg;
	
	// ++++++++++++++++++++++++++ Leuchtdioden + Schalter ++++++++++++++++++++++++++

	always @(posedge BCLK)
		if (WRSEL[5] && (ADDR[4:2] == 3'd1))
		  begin
			if (BE[3]) ss_reg[27:21] <= ~DIN[30:24];	// "1" = leuchten
			if (BE[2]) ss_reg[20:14] <= ~DIN[22:16];
			if (BE[1]) ss_reg[13: 7] <= ~DIN[14:8];
			if (BE[0]) ss_reg[ 6: 0] <= ~DIN[6:0];
		  end
		  else
			if (WRSEL[5] && (ADDR[4:2] == 3'd2))
			 begin
			   if (BE[1]) ss_reg[41:35] <= ~DIN[14:8];
			   if (BE[0]) ss_reg[34:28] <= ~DIN[6:0];	// Extension Register
			 end
		  
	always @(posedge BCLK) if (WRSEL[5] && (ADDR[4:2] == 3'd0) && BE[1]) LEDR[9:8] <= DIN[9:8];
	always @(posedge BCLK) if (WRSEL[5] && (ADDR[4:2] == 3'd0) && BE[0]) LEDR[7:0] <= DIN[7:0];

	// ++++++++++++++++++++++++++ Halt Monitor +++++++++++++++++++++++++
	
	always @(posedge BCLK) halt_cnt <= (READY | ~access) ? 20'd0 : halt_cnt + {19'd0,inc_halt_cnt};

	always @(posedge BCLK) inc_halt_cnt <= (halt_cnt[19:1] != 19'h7FFFF);
	
	//	Segment LSB ist oben, Segment MSB Strich in der Mitte:				H		A	L		T				S	t		o	p
	always @(posedge BCLK) SSEG <= inc_halt_cnt ? ss_reg : (READ ? ~{14'd0,7'h76,7'h77,7'h38,7'h78} : ~{14'd0,7'h6D,7'h78,7'h63,7'h73});
	
	// ++++++++++++++++++++++++++ Debug ++++++++++++++++++++++++++++++++
	
	// ++++++++++++++++++++++++++ NMI + Counter ++++++++++++++++++++++++
	
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) counter <= 32'd0;
		  else counter <= counter + 32'd1;	// Counter kann nicht angehalten werden
		  
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) tick_c <= 10'd0;
		  else if (tick) tick_c <= second ? 10'd0 : tick_c + 10'd1;
	
	always @(posedge BCLK) old_c15 <= counter[15];
	assign second = (tick_c == 10'd762);	// 763 * 65.536 = 50.003.968 langsamer als eine Sekunde...
	assign tick	  = ~counter[15] & old_c15;
	assign sec_puls = tick & second;		// fuer Laufschrift und auch fuer M3001 in Ceres
	
	always @(posedge BCLK) nmi_puls <= (tick & second) | (nmi_puls & ~counter[4]);	// 16 Takte
	
	always @(posedge BCLK) NMI_N <= ~(nmi_puls & nmie);	// einmal pro Sekunde

	assign wr_coun = WRSEL[5] & (ADDR[4:2] == 3'd3) & BE[3];

	always @(posedge BCLK or negedge BRESET)	// NMI Enable
		if (!BRESET) nmie <= 1'b0;
		  else if (wr_coun) nmie <= DIN[31];	// ist SET faehig
		  
	// ++++++++++++++++++++++++++ RESET Signal ++++++++++++++++++++++++++
	// Achtung : Einbau eines zweiten Reset-Signals fuer die PLL ! Es kommt 8 Takte frueher als der System-Reset.
	// Aenderung vom 8. Oktober 2011 im Rahmen von ModelSim Umstellung
	
	always @(posedge BCLK or negedge RST_N)
		if (!RST_N) init_cou <= 4'h0;
		  else init_cou <= init_cou + 4'h1;
		  
	always @(posedge BCLK or negedge RST_N)
		if (!RST_N) BRESET <= 1'b0;
		  else
			if (init_cou == 4'hF) BRESET <= 1'b1;

	// ++++++++++++++++++++++++ Steuerung HEX_OUT & DUAL_MODE neu am 19.6.2022 +++++++++++++++++++++++++
	
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) hex_out <= 1'b0;
		  else
			if (WRSEL[15] && BE[3] && (DIN[31:24] == 8'h71)) hex_out <= 1'b1;	// nur abschaltbar durch BRESET - spielen macht kein Sinn !

	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) dual_mode <= 1'b0;
		  else
			if (WRSEL[15] && BE[3] && (DIN[31:24] == 8'h3B)) dual_mode <= 1'b1;	// nur abschaltbar durch BRESET - spielen macht kein Sinn !
			
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) obe_ver <= 1'b0;
		  else
			if (WRSEL[15] && BE[3] && ((DIN[31:24] == 8'h17) | (DIN[31:24] == 8'hA5))) obe_ver <= 1'b1;	// nur abschaltbar durch BRESET !
			
	always @(posedge BCLK or negedge BRESET)
		if (!BRESET) ceres <= 1'b0;
		  else
			if (WRSEL[15] && BE[3] && (DIN[31:24] == 8'hA5)) ceres <= 1'b1;	// nur abschaltbar durch BRESET - spielen macht kein Sinn !
			
	assign CONFIG = {sec_puls,ceres,obe_ver,dual_mode,hex_out};
	
	assign DRAMSZ = 3'b011;	// 64MB

endmodule
